@@ -163,8 +163,6 @@ gtk_clutter_actor_unrealize (ClutterActor *actor)
   if (priv->widget == NULL)
     return;
 
-  gtk_widget_unrealize (priv->widget);
-
   priv->surface = NULL;
 
   g_object_ref (priv->widget);
@@ -511,6 +509,8 @@ gtk_clutter_actor_init (GtkClutterActor *self)
   g_object_ref_sink (priv->widget);
   gtk_widget_show (priv->widget);
 
+  clutter_actor_set_reactive (actor, TRUE);
+
   clutter_actor_push_internal (actor);
 
 #if defined(CLUTTER_WINDOWING_X11)
